/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/**
 * @file co_exception.c
 * @brief Cleanup stack exception handling implementation.
 **/
/* }}} */

/* {{{ includes */
#include <stdlib.h>
#include <stdio.h>
#include "co_exception.h"
#include "co_debug.h"
#include "co_util.h"
/* }}} */

#include "co_exception.cometa"

//cox_state *(*__cox_state) (void);
//static cox_state *__cox_state_simple;
cox_state *__cox_state;

//static cox_state *cox_state_simple (void);
static cox_state *cox_state_new (void);

/* {{{ */
void cox_stack_grow (void)
	/// Allocate more memory for cleanup stack.
{
	cox_stack *st;
	CO_DEBUG (8, "growing cleanup stack");
	st = malloc (sizeof (cox_stack));
	if (!st) COX_THROW (nomem);
	st -> n = __cox_state -> stack;
	__cox_state -> stack = st;
	__cox_state -> stack_size += _COX_STACK_SIZE;
}
/* }}} */
/* {{{ */
void cox_stack_shrink (void)
	/// Free unused memory from cleanup stack.
{
	cox_stack *st;
	CO_DEBUG (8, "shrinking cleanup stack");
	st = __cox_state -> stack;
	__cox_state -> stack = st -> n;
	free (st);
	__cox_state -> stack_size -= _COX_STACK_SIZE;
}
/* }}} */
/* {{{ */
void cox_unwind (unsigned mark)
	/// Remove all items off top of stack up to @a mark.
{
	unsigned idx; cox_stack_it *item;
	int i = __cox_state -> stack_top - mark;
	CO_DEBUG (8, "topmost handler: %p (mark = %d)", __cox_state -> cur_h,
			__cox_state -> cur_h ?
			__cox_state -> cur_h -> stack_mark : -1);
	CO_DEBUG (8, "unwinding last %d item(s)", i);
	if (i < 0) abort (); // FIXME?
	for (; i > 0; i --) {
		assert (__cox_state -> stack_top != 0);
		__cox_state -> stack_top --;
		idx = __cox_state -> stack_top % _COX_STACK_SIZE;
		item = &(__cox_state -> stack -> items [idx]);
		CO_DEBUG (1, "unwinding: item %p with idx %d", item -> item, idx);
		item -> cleanup (item -> item, item -> cleanup_data);
		if (! idx && __cox_state -> stack_size > _COX_STACK_SIZE)
			cox_stack_shrink ();
	}
}
/* }}} */

/* {{{ */
cox_state *cox_state_new (void)
	/// Create new CSEH state object.
{
	cox_state *s;
	s = malloc (sizeof (cox_state));
	s -> stack = malloc (sizeof (cox_stack));
	if (! s || ! s -> stack) {
		// FIXME!
		fprintf (stderr, "\nout of memory!\n");
		exit (1);
	}
	s -> stack_size = _COX_STACK_SIZE;
	s -> stack_top = 0;
	s -> cur_h = 0;
	return s;
}
/* }}} */
/* {{{ */
void cox_state_free (cox_state *s)
	/// Create new CSEH state object.
{
	free (s -> stack);
	free (s);
}
/* }}} */
/* {{{ */
void __co_exception_init (void)
	/// Initialize CSEH subsystem.
{
	//cox_state *s;

	__cox_state = cox_state_new ();
	//__cox_state_simple = s;
	//__cox_state = cox_state_simple;
}
/* }}} */
/* {{{ */
void __co_exception_fini (void)
	/// Initialize CSEH subsystem.
{
	cox_unwind (0);
	cox_state_free (__cox_state);
}
/* }}} */

/* {{{ */
void __cox_throw (long long err)
	/// Throw exception.
{
	cox_handler *h;
	h = __cox_state -> cur_h;
	if (! h) return;
	h -> error = err;
	cox_unwind (h -> stack_mark);
	longjmp (h -> jb, h -> error);
}
/* }}} */


/* {{{ */
void *cox_malloc (size_t size)
	/// Allocate memory and push it onto cleanup stack.
	/**
	 * This also throws exception if program runs out of memory. Uses malloc()
	 * to get memory and free() to free it.
	 * @param size - size of block to be allocated.
	 **/
{
	void *p;
#if RANDOM_MEM_EXHAUSTION
	if (rand () < RAND_MAX / 500) COX_THROW (nomem);
#endif
	p = malloc (size);
	if (!p) {
		CO_DEBUG (3, "failed to allocate %u bytes of memory", size);
		COX_THROW (nomem);
	}
	cox_push (p, (cox_cleanup_f) free, 0);
	memset (p, 0, size);
	return p;
}
/* }}} */
/* {{{ co_*_new helper macros */
#define CO_ARRAY_ADM(last, ar, ar_s, type) \
	do { \
		if (! ar) { \
			ar_s = 30; \
			ar = new_ar_p (type, ar_s); \
		} \
		if (ar_s <= last) { \
			ar_s += 10; \
			ar = realloc (ar, sizeof (type) * ar_s); \
			memset (ar + ar_s - 10, 0, 10 * sizeof (type)); \
		} \
	} while (0)
/* }}} */

#if 0
/* {{{ */
cox_state *cox_state_simple (void)
	/// Get CSEH state object, function for non-threaded programs.
{
	return __cox_state_simple;
}
/* }}} */
/* {{{ */
const char *co_error_string (int id)
{
	return __exception_id_string [id];
}
/* }}} */
static const char **__exception_id_string = 0;
static int __exception_id_string_sz = 0;
/* {{{ */
int __co_exception_id_new (const char *descr)
{
	static int last = 1; // 0 is ERR_OK
	CO_DEBUG (2, "adding exception %d: %s", last, descr);
	CO_ARRAY_ADM (last, __exception_id_string,
			__exception_id_string_sz, char *);
	__exception_id_string [last] = descr;
	CO_DEBUG (2, "result: %s", co_error_string (last));
	++ last;
	return last - 1;
}
/* }}} */
/* {{{ */ static inline void *cox_realloc (void *p, size_t size)
	/// DO NOT USE!
{
	cox_pop (p);
	p = realloc (p, size);
	if (!p) COX_THROW (nomem);
	cox_push (p, (cox_cleanup_f)free, 0);
	return p;
} /* }}} */
#endif
