/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/** @file co_object.c
 *  @brief Object definition, creation and manipulation. */
/* }}} */

#define I18N_NOOP(x)
I18N_NOOP ("test");
#undef I18N_NOOP

/* {{{ includes */
#include "co_util.h"
#include "co_gettext.h"
#include "co_object.h"
#include "co_exception.h"
#include "co_conv.h"
/* }}} */

/* {{{ type database *//*
struct co_o_type *__co_type_array = 0;
unsigned __co_type_array_size = 0;
char **__co_action_array = 0;
unsigned __co_action_array_size = 0;
char **__co_event_array = 0;
unsigned __co_event_array_size = 0;
char **__co_property_array = 0;
unsigned __co_property_array_size = 0;
*//* }}} */

#include "co_object.cometa"

/* {{{ */
static co_v co_msg_default (co this, co self, co_id_t message, va_list ap)
	/// Default, table-driven message handler for object (internal).
	/**
	 * This function shouldn't be used in any explicit way. It will be
	 * assigned to objects which do not implement it's own message-arrival
	 * handler automatically.
	 *
	 * @internal
	 **/
{
	co_o_head_full head = this;
	co_meth_entry def = 0;
	co_meth_entry cur = 0;
	unsigned i;

	for (i = 0; i < head -> __co_meth_ar -> n; i ++) {
		cur = head -> __co_meth_ar -> meth + i;
		CO_DEBUG (3, "trying handler %d == message %d",
				cur -> message, message);
		if (message == cur -> message)
			return cur -> handler (this, self, message, ap);
		if (cur -> message == COM (default))
			def = cur;
	}

	// fallback to default method
	if (def)
		return def -> handler (this, self, message, ap);

	COX_THROW (message_unhandled);
	// just to silence compiler
	CO_RETURN (int, 0xBAD0);
}
/* }}} */

/* {{{ */
co_v co_clone_shallow (co this, co self, co_id_t message, va_list ap)
{
	//co_o_head_full head = this;
	(void) message; (void) self; (void) ap;
	CO_RETURN (co, this);
}
/* }}} */
/* {{{ */
co_v co_clone_deep (co this, co self, co_id_t message, va_list ap)
{
	co n = co_clone_shallow (this, self, message, ap) . _co;
	M (n, copy_deep, this); // ask object to copy over its data
	CO_RETURN (co, n);
}
/* }}} */
/* {{{ */
co_v co_clone_full (co this, co self, co_id_t message, va_list ap)
{
	co n = co_clone_deep (this, self, message, ap) . _co;
	// TODO: copy the method table
	CO_RETURN (co, n);
}
/* }}} */
/* {{{ */
co_v co_clone_clean (co this, co self, co_id_t message, va_list ap)
{
	co n = co_clone_shallow (this, self, message, ap) . _co;
	CO_RETURN (co, n);
}
/* }}} */
/* {{{ */
co_v co_clone_clean_full (co this, co self, co_id_t message, va_list ap)
{
	co n = co_clone_shallow (this, self, message, ap) . _co;
	// TODO: copy the method table
	CO_RETURN (co, n);
}
/* }}} */

/* {{{ */
static void co_listen_cleanup (co o)
	/// Clean up listener list (internal).
	/**
	 * This function is called automatically and shouldn't be used directly.
	 * It cleans up the list of event listeners, so we regain memory and speed
	 * up lookup in co_ev function.
	 **/
{
	co_o_head_full _o = o;
	co_ev_l i, in;

	CO_DEBUG (3, "removing from head...");
	while (_o -> __co_listen && _o -> __co_listen -> deleted) {
		i = _o -> __co_listen;
		_o -> __co_listen = i -> n;
		free (i);
	}

	CO_DEBUG (4, "removing from tail...");
	for (i = _o -> __co_listen; i && i -> n; i = i -> n) {
		if (i -> n -> deleted) {
			in = i -> n;
			i -> n = in -> n;
			free (in);
		}
	}
}
/* }}} */
/* {{{ */
static void co_o_disconnect (co _emit, co listen)
	/// Disconnect two objects.
	/**
	 * This removes all references to @a listen from @a _emit. When @a listen
	 * is being destroyed, this is called automatically.
	 **/
{
	co_o_head_full emit;
	co_ev_l i;

	emit = _emit;

	CO_DEBUG (4, "marking items for deletion (_o = %p, dis = %p)...",
			emit, listen);
	for (i = emit -> __co_listen; i; i = i -> n) {
		CO_DEBUG (8, "testing: i = %p, i -> o = %p", i, i -> o);
		if (i -> o == listen) {
			CO_DEBUG (6, "marked: %p", i);
			i -> deleted = 1;
			emit -> check_listen = 1;
		}
	}
}
/* }}} */

/* {{{ */
static void co_o_destroy (co o)
	/// Destroy object (internal: no checking is done).
	/**
	 * This destroys object without checking the state prior to taking
	 * actions and without calling object's own destruction handler. Do not
	 * use or if absolutely needed, only with extreme caution. Use co_del_f
	 * for standard but forced destruction or co_del for standard checked
	 * destruction.
	 *
	 * @see co_del, co_del_f
	 **/
{
	co_o_head_full _o = o;
	co_ev_l i, in;
	CO_DEBUG (4, "cleaning _o (= %p) -> __co_listen", _o);
	for (i = _o -> __co_listen; i; i = i -> n) {
		CO_DEBUG (6, "disconnecting (i = %p, deleted = %d)", i, i -> deleted);
		if (i -> ev == COE (disconnect_object) && ! i -> deleted)
			co_o_disconnect (i -> o, o);
	}
	for (i = _o -> __co_listen; i; i = in) {
		in = i -> n;
		free (i);
	}
}
/* }}} */
/* {{{ */
static void __co_del (co o)
	/// Delete object (internal).
	/**
	 * Is identical to co_del_f, but internal. Use co_del_f instead.
	 **/
{
	CO_DEBUG (3, "performing object destruction: %p", o);
	((co_o_head) o) -> destroy = 1;
	CO_EV (o, destroyed);
	co_o_destroy (o);
	((co_o_head_full) o) -> __co_msg (o, o, COM (destroy), 0);
} /* }}} */
/* {{{ */
void co_del (co o)
{
	if (((co_o_head) o) -> destroy) return; // already being destroyed :)
	if (((co_o_head) o) -> msg_proc) {
		((co_o_head) o) -> destroy = 1;
		CO_DEBUG (3, "postponed object deletion: %p", o);
	} else {
		CO_DEBUG (3, "immediate object deletion: %p", o);
		__co_del (o);
	}
} /* }}} */
/* {{{ */
void co_del_f (co o)
{
	CO_DEBUG (2, "forcing object deletion: %p", o);
	__co_del (o);
} /* }}} */

/* {{{ */
void co_bless (co o, co_id_t type, co_msg_h msg)
{
	co_o_head_full _o = o;

	_o -> type = type;
	_o -> __co_msg = 0;
	_o -> destroy = 0;
	//_o -> __co_prop = p ? p : co_prop_min;
	_o -> __co_msg = msg ? msg : co_msg_default;
	_o -> __co_listen = 0;
	_o -> ev_running = 0;
	_o -> check_listen = 0;
}
/* }}} */

/* {{{ */
static void co_m_v_recov (void *o, const void *ignore)
{
	(void) ignore;
	((co_o_head_full) o) -> msg_proc --;
}
/* }}} */
/* {{{ */
co_v co_m_v (co o, co s, co_id_t a, va_list ap)
{
	co_o_head_full _o = o;
	int db;
	volatile co_v rv;
#if DEBUGLVL > 0
	char __msg_id [128];
	int __lastdbg = __co_debug_msg_level;
#endif

	db = _o -> destroy;
	_o -> msg_proc ++;
	cox_push (_o, co_m_v_recov, 0);
	CO_DEBUG (1, "processing: o = %p, s = %p, msg = %d", o, s, a);
/* {{{ debug */
#if DEBUGLVL > 0
	__co_debug_msg_level ++;
	sprintf (__msg_id, "unknown %d (FIXME)", a);
	//snprintf (__msg_id, 128, "%s/%s",
	//		__co_type_array [_o -> t] . name, __co_action_array [a]);
	CO_DEBUG2 (2, "---------------------------------------------------"
			"-----------------------------\n");
	CO_DEBUG2 (2, "   BEGIN HANDLE MESSAGE | %23s | %p\n", __msg_id, o);
#endif
/* }}} */
	rv = _o -> __co_msg (o, s, a, ap);
/* {{{ debug */
#if DEBUGLVL > 0
	CO_DEBUG2 (2, "     END HANDLE MESSAGE | %23s | %p\n", __msg_id, o);
	CO_DEBUG2 (2, "---------------------------------------------------"
			"-----------------------------\n");
	__co_debug_msg_level --;
#endif
/* }}} */
	_o -> msg_proc --;
	cox_pop (_o); // FIXME? this might be a race condition
	if ((!db) && (! (_o -> msg_proc)) && _o -> destroy)
		__co_del (o);

	return rv;
} /* }}} */
/* {{{ */
co_v co_m (co o, co s, co_id_t a, ...)
{
	va_list ap;
	co_v rv;

	va_start (ap, a);
	rv = co_m_v (o, s, a, ap);
	va_end (ap);

	return rv;
} /* }}} */
/* {{{ */
void co_ev (co o, co_id_t ev, ...)
{
	volatile co_ev_l i;
	co_id_t ot;
	int erb;
	co_o_head_full _o = o;
	va_list ap;
#if DEBUGLVL > 0
	int __dbglvl = __co_debug_msg_level;
	//char __ev_id [128];
#endif

	erb = _o -> ev_running;
	_o -> ev_running = 1;

	ot = _o -> type;
/* {{{ debug *//*
#if DEBUGLVL > 0
	__co_debug_msg_level ++;
	snprintf (__ev_id, 128, "%s/%s",
			__co_type_array [_o -> type] . name, __co_event_array [ev]);
	CO_DEBUG2 (2, "---------------------------------------------------"
			"-----------------------------\n");
	CO_DEBUG2 (2, "            BEGIN EVENT | %23s | %p\n", __ev_id, o);
#endif
*//* }}} */
	for (i = _o -> __co_listen; i; i = i -> n) {
		if (ev == i -> ev && ! i -> deleted) {
			//CO_DEBUG (2, "match: o = %p, msg = %s", i -> o, __co_action_array [i -> msg]);
			va_start (ap, ev);
			if (i -> msg == COM (destroy)) {
				i -> deleted = 1;
				co_del (i -> o);
			} else {
				if (! erb) {
					COX_TRY {
						co_m_v (i -> o, i -> o, i -> msg, ap);
					}
					COX_CATCH (err) {
						_o -> ev_running = 0;
#if DEBUGLVL > 0
						__co_debug_msg_level = __dbglvl;
#endif
						COX_RETHROW (err);
					} COX_CATCH_END;
				} else {
					co_m_v (i -> o, i -> o, i -> msg, ap);
				}
			}
			va_end (ap);
			if (i -> ad) i -> deleted = 1;
		}
	}
/* {{{ debug *//*
#if DEBUGLVL > 0
	CO_DEBUG2 (2, "              END EVENT | %23s | %p\n", __ev_id, o);
	CO_DEBUG2 (2, "---------------------------------------------------"
			"-----------------------------\n");
	__co_debug_msg_level --;
#endif
*//* }}} */
	_o -> ev_running = erb;
	if (! erb & _o -> check_listen)
		co_listen_cleanup (o);
} /* }}} */

/* {{{ */
void co_cnt (co emit, co_id_t ev, co __co_listen, co_id_t msg, bool ad)
{
	co_ev_l i, nw;
	co_o_head_full _e = emit;

	/* CO_DEBUG (2, "%p (%s) listens to %p (%s))",
			__co_listen, __co_type_array [((co_o) __co_listen) -> type] . name,
			emit, __co_type_array [_e -> type] . name);
	CO_DEBUG (2, "msg = %s, ev = %s",
		__co_action_array [msg],
		__co_event_array [ev]); */

	// do not create two identical connections
	for (i = _e -> __co_listen; i; i = i -> n) {
		if (i -> ev == ev && __co_listen == i -> o && i -> msg == msg) {
			CO_DEBUG (2, "already exists, aborting");
			return;
		}
	}

	CO_DEBUG (3, "creating connection data...");
	nw = new_p (co_ev_l_s);
	nw -> o = __co_listen;
	nw -> ev = ev;
	nw -> msg = msg;
	nw -> n = 0;
	nw -> deleted = 0;
	nw -> ad = ad;

	CO_DEBUG (3, "embedding data into linked list...");
	if (_e -> __co_listen) {
		nw -> n = _e -> __co_listen;
	}
	_e -> __co_listen = nw;

	if (emit != __co_listen) {
		CO_DEBUG (3, "creating auto-disconnect connection...");
		CO_CNT (__co_listen, disconnect_object, emit, disconnect_object);
	}
} /* }}} */

/* {{{ */
co __co_o_bless (co o)
{
	CO_DEBUG (3, "blessing object %p", o);
	CO_PUSH (o);
	CO_M (o, bless);
	return o;
}
/* }}} */
/* {{{ */
co __co_o_bless_p (co o)
{
	CO_DEBUG (3, "blessing object %p", o);
	CO_M (o, bless);
	return o;
}
/* }}} */

/* {{{ */
co_v co_has_method (co this, co self, co_id_t message, va_list ap)
	/// Reflection message handler.
{
	co_o_head_full head = this;
	co_id_t find = va_arg (ap, co_id_t);
	unsigned i;
	(void) message; (void) self;

	CO_DEBUG (3, "co_has_method called");

	for (i = 0; i < head -> __co_meth_ar -> n; i ++) {
		// cur = head -> __co_meth_ar -> meth + i;
		if (find == head -> __co_meth_ar -> meth [i] . message) {
			CO_DEBUG (3, "found[%d]: %d", i, find);
			CO_RETURN (int, 1);
		}
	}

	CO_RETURN (int, 0);
}
/* }}} */
/* {{{ */
void *co_meth_array_new (void)
{
	co_meth_array ar;
	ar = cox_malloc (sizeof (co_meth_array_s));
	ar -> n = 0;
	ar -> size = 4;
	ar -> meth = cox_malloc (sizeof (co_meth_entry_s) * 4);
	cox_pop (ar -> meth);
	cox_pop (ar);
	return ar;
}
/* }}} */
/* {{{ */
co_id_t co_meth_add (co_meth_array ar, co_id_t msg, co_msg_h handl)
{
	if (ar -> n >= ar -> size) {
		CO_DEBUG (2, "growing %p", ar);
		ar -> size *= 2;
		ar -> meth = realloc (ar -> meth, ar -> size *
				sizeof (co_meth_entry_s));
	}
	ar -> meth [ar -> n] . message = msg;
	ar -> meth [ar -> n] . handler = handl;
	//ar -> master [ar -> n] . name = name;
	//co_meth_sort (ar);
	ar -> n ++;
	return msg;
}
/* }}} */
/* {{{ */
void co_meth_array_free (co_meth_array ar)
{
	free (ar -> meth);
	free (ar);
}
/* }}} */
