/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002, 2003                                              *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/** @file co_object.h
 *  @brief Object definition and function prototypes. */
/* }}} */

#ifndef __CO_OBJECT_H__
#define __CO_OBJECT_H__

#include <stdarg.h>
#include "co_type.h"
#include "co_exception.h"
#include "co_string.h"
#include "co_debug.h"

#if DEBUGLVL > 0
#include <stdio.h>
#endif

/* {{{ co message and event macros */
/// Send message to object @a o (message selector is first arg).
#define CO_M(o, ...) (CO_DEBUG (8, "sending msg..."), \
		co_m (o, o, COM (__VA_ARGS__)))
#define CO_D(o, ...) (co_m (o, this, COM (__VA_ARGS__)))
#define CO_D_V(o, s, m, v) (co_m_v (o, s, m, v))
#define CO_A(t, o, ...) (CO_DEBUG (8, "getting attribute..."), \
		* (co_m (o, o, COM (__VA_ARGS__)) . _ ## t ## _p))
/// Send message to object @a o (message selector is first arg).
#define CO_M_V(o, a, va) (CO_DEBUG (8, "sending msg..."), \
		co_m_v (o, COM (a), va))

/// Raise event in object @a o.
#define CO_EV(o, ...) (co_ev (o, COE (__VA_ARGS__)))

/// Connect object @a a event (signal) to object @a b message (slot)
#define CO_CNT(a, e, b, m) (co_cnt ((a), (COE (e)), b, COM (m), false))
#define CO_CNT_AD(a, e, b, m) (co_cnt ((a), (COE (e)), b, COM (m), true))
/* }}} */
/* {{{ co creation macros */
#define CO_BLESS(o, t, m) (co_bless ((o), COT (t), (m)))
#define CO_BLESS_0(o, t) CO_BLESS (o, t, 0)
#define CO_BLESS_M(o, t) CO_BLESS (o, t, t ## _msg)
#define CO_BLESS_P(o, t) CO_BLESS (o, t, 0)
#define CO_BLESS_MP(o, t) CO_BLESS (o, t, t ## _msg, t ## _p)
/// Create and bless object.
#define CO_NEW(t, ...) (__co_o_bless (t ## _new (__VA_ARGS__)))
#define CO_NEW_P(t, ...) (__co_o_bless_p (t ## _new (__VA_ARGS__)))
#define CO_F_NEW(t, ...) (__co_o_bless ((co_f_get (t)) (__VA_ARGS__)))
#define CO_F_NEW_P(t, ...) (__co_o_bless_p ((co_f_get (t)) (__VA_ARGS__)))
#define CO_FS_NEW(t, ...) (__co_o_bless ((co_fs_get (t)) (__VA_ARGS__)))
/* }}} */
/* {{{ co id construction macros */
/// Construct real co action identifier.
#define COM(...) __co_id_message_ ## __VA_ARGS__
/// Construct real co type identifier.
#define COT(...) __co_id_type_ ## __VA_ARGS__
/// Construct real co event identifier.
#define COE(...) __co_id_event_ ## __VA_ARGS__
/// Construct real co property identifier.
#define COP(...) __co_id_property_ ## __VA_ARGS__
/* }}} */
/* {{{ message handler helper macros */
/// Begin message handler switch.
#define CO_MH_BEGIN \
	CO_DEBUG (18, "entering message handler switch..."); \
	if (0)

/// Message handler switch case/option.
#define CO_MH_CASE(a, x) \
	else if (a == COM (x) \
			&& (CO_DEBUG (18, "chosen case: %s", #x), 1))

/// Message handler default switch case (MUST be last)
#define CO_MH_DEFAULT else

/* #define CO_MH_UNHANDLED(x) else (CO_DEBUG (1, \
			"WARNING: unhandled message %s", \
			__co_action_array [x])) */
#define CO_MH_UNHANDLED(x) else

/// End message handler switch (optional, IMHO).
#define CO_MH_END
/* }}} */

#define CO_PUSH(o) (cox_push ((o), (cox_cleanup_f) co_del_f, 0))
/*#define CO_OBJECT_IMPL(x, y)
#define CO_MSG_FW(x, y)
#define CO_MSG_FW_DEFAULT(x, y)*/

/* {{{ CO_POSSIBLY_UNUSED */
#ifdef __GNUC__
#define CO_POSSIBLY_UNUSED __attribute__ ((unused))
#else
#define CO_POSSIBLY_UNUSED
#endif
/* }}} */

#define CO_METH(x, y) co_v x ## _ ## y \
	(x this CO_POSSIBLY_UNUSED, co self CO_POSSIBLY_UNUSED, \
	 co_id_t message CO_POSSIBLY_UNUSED, va_list ap CO_POSSIBLY_UNUSED)
#define CO_METH_INI(x, y,  ...) CO_METH_INI_ ## x ## _ ## y
#define CO_OBJECT(x) struct x; CO_TYPEDEF (x); struct x
#define CO_ATTR(x, y) x y;

struct co;
typedef unsigned co_id_t;

typedef co (*co_constructor) ();

/// Function pointer to object message handler.
typedef co_v (*co_msg_h) (co o, co s, co_id_t a, va_list ap);
//typedef co_vr (*co_prop_h) (const co o, co_id_t p);

/* {{{ */
struct co_meth_entry
	/// Method table entry.
{
	co_id_t message;
	co_msg_h handler;
};
CO_TYPEDEF (co_meth_entry);
/* }}} */
/* {{{ */
struct co_meth_array
	/// Method table.
{
	size_t size;
	size_t n;
	co_meth_entry meth;
};
CO_TYPEDEF (co_meth_array);
/* }}} */

/* {{{ */
struct co_o_head
	/// Arbitrary object head - minimal version.
{
	/// Point to type_rec (which in turn may contain class reference)
	co_id_t type;
	/// Is set to 1 if object is being destroyed (internal)
	int destroy:1;
	/// Depth counter for message handler (internal)
	unsigned msg_proc:17;
	/// Is set to 1 if list of listening objects should be sweeped.
	int check_listen:1;
	/// Is set to 1 if event is being processed.
	int ev_running:1;
};
CO_TYPEDEF (co_o_head);
/* }}} */
/* {{{ object head #define's */
#define CO_HEAD_MIN \
	co_o_head_s __co_root
#define CO_HEAD_SA \
	co_o_head_s __co_root; \
	co_msg_h __co_msg
#define CO_HEAD_FULL \
	co_o_head_s __co_root; \
	co_msg_h __co_msg; \
	co_meth_array __co_meth_ar; \
	struct co_ev_l *__co_listen
/* }}} */
/* {{{ */
struct co_o_head_full
	/// Object head - full version.
{
	/// Point to type_rec (which in turn may contain class reference)
	co_id_t type;
	/// Is set to 1 if object is being destroyed (internal)
	int destroy:1;
	/// Is set to 1 if object is processing message right now (internal)
	int msg_proc:1;
	/// Is set to 1 if list of listening objects should be sweeped.
	int check_listen:1;
	/// Is set to 1 if event is being processed.
	int ev_running:1;
	co_msg_h __co_msg;
	co_meth_array __co_meth_ar;
	struct co_ev_l *__co_listen;
};
CO_TYPEDEF (co_o_head_full);
/* }}} */
/* {{{ */ struct co_ev_l
	/// Object connection record - event listener (internal).
{
	/// Message ID to send if activated.
	co_id_t msg;
	/// Event ID, if raised connection is activated.
	co_id_t ev;
	/// Object to send message to.
	co o;
	/// Next list member.
	struct co_ev_l *n;
	int deleted:1;
	int ad:1;
};
CO_TYPEDEF (co_ev_l); /* }}} */

/// Create new method table (internal).
/** @internal */
void *co_meth_array_new (void);

void co_meth_array_free (co_meth_array ar);

/// Add new item to (pre-existing) method table.
co_id_t co_meth_add (co_meth_array ar, co_id_t message, co_msg_h hendl);

/// Check if method is present in object.
co_v co_has_method (co o, co s, co_id_t a, va_list ap);

/// Send message @a a to object @a o, vararg version.
co_v co_m (co o, co s, co_id_t a, ...);

/// Send message to a object, va_list version.
/**
 * This function implements the message sending apparatus of libco. You
 * probably won't use it directly in most cases (see CO_M and co_m these
 * are used much more often). Only case when this function is directly
 * useful is when you forward messages (one object receives message and
 * sends it as-is to another object). Or in your own message sending
 * function.
 *
 * @param o - this is object that the message is sent to
 * @param a - this is the action id of this message (see COM and CO_M)
 * @param ap - variable arguments (passed to message handler intact)
 *
 * @return return value of message handler (depends on object)
 **/
co_v co_m_v (co o, co s, co_id_t a, va_list ap);

/// Raise event from object.
/**
 * This function is responsible for raising events. When event is invoked,
 * the object (in fact this function) will iterate over list of registered
 * listeners and send them messages as given by registration (arguments
 * are taken from event invoke handler). Currently there is no way to
 * predefine arguments to pass to message at registration time.
 *
 * @param o - object that raises event
 * @param ev - event id to check for handlers
 * @param varargs - arguments to pass to handlers
 **/
void co_ev (co o, co_id_t ev, ...);

//void co_o_destroy (co o);
//void __co_del (co o);

/// Delete object in a safe way.
/**
 * This function schedules object for destruction. This means: if object
 * still performs some actions, it will be destroyed when last one is
 * finished. Otherwise (the object is idle) it will be destroyed
 * immediately. This way, you can safely destroy objects from their
 * message handlers, without causing kind of lossage free()'ing them would
 * cause (You can't free them, because upon return from message handler
 * the state of object is manipulated - and it isn't wise to write into
 * free'd memory).
 *
 * @param o - object to delete
 **/
void co_del (co o);

/// Force immediate (but mostly clean) object deletion.
/**
 * This will free the object (cleanly), but won't check for ongoing
 * activity. <b>Don't use this function unless you are really sure that
 * object in question will never regain control.</b> This is useful in eg.
 * application object, where it destroys itself and then calls exit(), but
 * only in case you are purist and don't want to have unfreed memory upon
 * exit :).
 *
 * @param o - object to delete
 **/
void co_del_f (co o);

/// Connect event @a ev of object @a listener to @a msg of @a emitor.
void co_cnt (co emitor, co_id_t ev, co listener, co_id_t msg, bool ad);

/// Create new object and initialize it from given arguments.
void co_bless (co o, co_id_t t, co_msg_h msg);

/// Helper function for CO_O_NEW macro (internal).
/**
 * @internal
 **/
co __co_o_bless (co o);
/// Helper function for CO_O_NEW macro (internal).
/**
 * @internal
 **/
co __co_o_bless_p (co o);

co_v co_clone_shallow (co this, co self, co_id_t message, va_list ap);
co_v co_clone_deep (co this, co self, co_id_t message, va_list ap);
co_v co_clone_full (co this, co self, co_id_t message, va_list ap);
co_v co_clone_clean (co this, co self, co_id_t message, va_list ap);
co_v co_clone_clean_full (co this, co self, co_id_t message, va_list ap);

#endif /* __CO_OBJECT_H__ */
